"""Parallel processing functions for distill and inference pipelines."""

import os
from inspect import getsource
from pathlib import Path
from typing import Callable, Dict, List

import pandas as pd
from tqdm import trange

from zeno.api import DistillReturn, ModelReturn, ZenoOptions
from zeno.classes.base import DataProcessingReturn, ZenoColumn, ZenoColumnType
from zeno.util import load_series


def predistill_data(
    fn: Callable[[pd.DataFrame, ZenoOptions], DistillReturn],
    column: ZenoColumn,
    options: ZenoOptions,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> List[DataProcessingReturn]:
    col_hash = str(column)

    # To prevent SettingWithCopyWarning
    col = df[col_hash].copy()

    save_path = Path(cache_path, col_hash + ".pickle")
    to_predict_indices = col.loc[pd.isna(col)].index

    if len(to_predict_indices) > 0:
        for i in trange(
            0,
            len(to_predict_indices),
            batch_size,
            desc="preprocessing " + fn.__name__,
            position=pos,
        ):
            out = fn(df.loc[to_predict_indices[i : i + batch_size]], options)
            col.loc[to_predict_indices[i : i + batch_size]] = out.distill_output
            col.to_pickle(str(save_path))

    return [DataProcessingReturn(column=column, output=col)]


def run_inference(
    fn: Callable[[str], Callable[[pd.DataFrame, ZenoOptions], ModelReturn]],
    options: ZenoOptions,
    model_name: str,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> List[DataProcessingReturn]:
    model_col_obj = ZenoColumn(
        column_type=ZenoColumnType.OUTPUT, name="output", model=model_name
    )
    embedding_col_obj = ZenoColumn(
        column_type=ZenoColumnType.EMBEDDING, name="embedding", model=model_name
    )
    model_hash = str(model_col_obj)
    embedding_hash = str(embedding_col_obj)
    model_col = df[model_hash].copy()
    embedding_col = df[embedding_hash].copy()

    model_save_path = Path(cache_path, model_hash + ".pickle")
    embedding_save_path = Path(cache_path, embedding_hash + ".pickle")

    to_predict_indices = model_col.loc[pd.isna(model_col)].index

    other_return_cols: Dict[str, ZenoColumn] = {}
    if len(to_predict_indices) > 0:
        model_fn = fn(model_name)

        # Make output folder if function uses output_path
        src = getsource(model_fn)
        if "output_path" in src:
            file_cache_path = os.path.join(cache_path, model_hash)
            os.makedirs(file_cache_path, exist_ok=True)
            options = options.copy(update={"output_path": file_cache_path})

        for i in trange(
            0,
            len(to_predict_indices),
            batch_size,
            desc="Inference on " + model_name,
            position=pos,
        ):
            out = model_fn(df.loc[to_predict_indices[i : i + batch_size]], options)

            # Check if we also get embedding
            if out.embedding is not None:
                for j, idx in enumerate(to_predict_indices[i : i + batch_size]):
                    model_col.at[idx] = out.model_output[j]  # noqa: PD008
                    embedding_col.at[idx] = out.embedding[j]  # noqa: PD008

                embedding_col.to_pickle(str(embedding_save_path))
            else:
                model_col.loc[to_predict_indices[i : i + batch_size]] = out.model_output

            if out.other_returns is not None:
                for k, v in out.other_returns.items():
                    if i == 0:
                        postdistill_col_obj = ZenoColumn(
                            column_type=ZenoColumnType.POSTDISTILL,
                            name=k,
                            model=model_name,
                        )
                        other_return_cols[k] = postdistill_col_obj
                        if str(postdistill_col_obj) not in df.columns:
                            load_series(
                                df,
                                postdistill_col_obj,
                                Path(cache_path, str(postdistill_col_obj) + ".pickle"),
                            )

                    postdistill_col_obj = other_return_cols[k]
                    postdistill_hash = str(postdistill_col_obj)
                    postdistill_col = df[postdistill_hash].copy(deep=False)
                    postdistill_col.loc[
                        to_predict_indices[i : i + batch_size]
                    ] = out.other_returns[k]
                    postdistill_save_path = Path(
                        cache_path, postdistill_hash + ".pickle"
                    )
                    postdistill_col.to_pickle(str(postdistill_save_path))

            model_col.to_pickle(str(model_save_path))

    ret = [DataProcessingReturn(column=model_col_obj, output=model_col)]
    if not embedding_col.isna().to_numpy().any():  # type: ignore
        ret.append(DataProcessingReturn(column=embedding_col_obj, output=embedding_col))
    for k, v in other_return_cols.items():
        ret.append(DataProcessingReturn(column=v, output=df[str(v)]))

    return ret


def postdistill_data(
    fn: Callable[[pd.DataFrame, ZenoOptions], DistillReturn],
    model: str,
    options: ZenoOptions,
    cache_path: str,
    df: pd.DataFrame,
    batch_size: int,
    pos: int,
) -> List[DataProcessingReturn]:
    col_obj = ZenoColumn(
        column_type=ZenoColumnType.POSTDISTILL,
        name=fn.__name__,
        model=model,
    )
    col_hash = str(col_obj)
    col = df[col_hash].copy()

    output_obj = ZenoColumn(
        column_type=ZenoColumnType.OUTPUT, name="output", model=model
    )
    output_hash = str(output_obj)

    save_path = Path(cache_path, col_hash + ".pickle")

    to_predict_indices = col.loc[pd.isna(col)].index

    local_options = options.copy(
        update={
            "output_column": output_hash,
            "output_path": os.path.join(cache_path, output_hash),
        }
    )

    if len(to_predict_indices) > 0:
        for i in trange(
            0,
            len(to_predict_indices),
            batch_size,
            desc="postprocessing " + fn.__name__ + " on " + model,
            position=pos,
        ):
            out = fn(df.loc[to_predict_indices[i : i + batch_size]], local_options)
            col.loc[to_predict_indices[i : i + batch_size]] = out.distill_output
            col.to_pickle(str(save_path))

    return [DataProcessingReturn(column=col_obj, output=col)]



"""The `zeno init` command, which sets up a new Zeno project."""

import os

import inquirer
import requests
from inquirer import Path, errors

from zeno.util import VIEW_MAP_URL, VIEWS_MAP_JSON

TOML_TEMPLATE = """functions = "./"
view = "{}"
metadata = "{}"
data_column = "{}"
{}
models = [] # Add paths to model checkpoints, passed into @model
"""


def validate_name(_, name):
    if len(name) == 0:
        raise errors.ValidationError("", reason="Project name cannot be empty.")
    return True


def setup_zeno():
    print(
        "Welcome to \033[1mZeno\033[0m! The following questions will guide "
        + "you through setting up your project.\n"
    )

    views_res = requests.get(VIEW_MAP_URL + VIEWS_MAP_JSON)
    views = list(views_res.json().keys())

    questions = [
        inquirer.Text(
            "name",
            message="Name your project/folder",
            validate=validate_name,
        ),
        inquirer.List(
            "view", message="What type of model are you evaluating", choices=views
        ),
        inquirer.Path(
            "metadata",
            message="Path to a metadata file (CSV or Parquet)",
            path_type=Path.FILE,
            exists=True,
        ),
        inquirer.Text(
            "data_column",
            message="Name of the column with data instances or paths"
            + "in the metadata file",
        ),
        inquirer.Text(
            "label_column",
            message="Name of the column with labels in the metadata file",
        ),
    ]

    result = inquirer.prompt(questions)
    if result is None:
        return

    print(
        "\nCongrats! A new folder, {}, has been created with a ".format(result["name"])
        + "zeno.toml file inside. Run Zeno with the following:"
    )
    print(("\n\t cd {} \n\t zeno zeno.toml").format(result["name"]))

    print(
        "\n If your data is saved as files, e.g. images or video, please set the ",
        "data_path option in the config file. To set up your model, follow the ",
        "instructions in model.py",
    )

    # Create new directory with zeno.toml file
    os.mkdir(str(result["name"]))

    # Update paths to be relative to new directory if relative to start with.
    if not os.path.isabs(str("metadata")):
        result["metadata"] = os.path.relpath(
            str(result["metadata"]), "./" + str(result["name"])
        )

    os.chdir(str(result["name"]))

    with open("zeno.toml", "w") as f:
        f.write(
            TOML_TEMPLATE.format(
                result["view"],
                result["metadata"],
                result["data_column"],
                'label_column = "' + str(result["label_column"]) + '"'
                if result["label_column"]
                else "",
            )
        )
    # Create model.py file



from zeno.api import (
    DistillReturn,
    MetricReturn,
    ModelReturn,
    ZenoOptions,
    ZenoParameters,
    distill,
    metric,
    model,
)
from zeno.runner import get_server, zeno

__all__ = [
    "DistillReturn",
    "MetricReturn",
    "ModelReturn",
    "ZenoOptions",
    "ZenoParameters",
    "get_server",
    "distill",
    "metric",
    "model",
    "zeno",
]



import secrets
from typing import List

import numpy as np
import pandas as pd
from sliceline.slicefinder import Slicefinder

from zeno.classes.base import MetadataType
from zeno.classes.slice import FilterPredicate, FilterPredicateGroup, Slice
from zeno.classes.slice_finder import SliceFinderRequest, SliceFinderReturn
from zeno.processing.filtering import filter_table
from zeno.util import generate_diff_cols


# Discretize continuous valued columns.
def cont_cols_df(df, cols: List[str]):
    new_df = pd.DataFrame()
    for col in cols:
        df_col = df.loc[:, col].copy()
        bins = list(np.histogram_bin_edges(df_col, bins="doane"))
        bins[0], bins[len(bins) - 1] = bins[0] - 1, bins[len(bins) - 1] + 1
        new_df.loc[:, col + "_encode"] = pd.cut(df_col, bins=bins)
    return new_df


def slice_finder(df, req: SliceFinderRequest):
    """Return slices of data with high or low metric values.

    Args:
        df (DataFrame): Zeno DataFrame with all metadata.
        req (SliceFinderRequest): Request with columns, metrics, and options.

    Returns a SliceFinderMetricReturn Object.
    """
    cont_search_cols, not_cont_search_cols = [], []
    for col in req.search_columns:
        if col.metadata_type == MetadataType.CONTINUOUS:
            cont_search_cols.append(col)
        else:
            not_cont_search_cols.append(col)

    search_cols = not_cont_search_cols + cont_search_cols
    cont_search_cols = [str(col) for col in cont_search_cols]
    not_cont_search_cols = [str(col) for col in not_cont_search_cols]
    metric_col = "diff" if req.compare_column else str(req.metric_column)

    filt_df = filter_table(
        df, req.filter_predicates, req.filter_ids, req.tag_ids, req.tag_list
    )
    cont_df = cont_cols_df(filt_df[cont_search_cols].dropna(), cont_search_cols)

    if req.compare_column:
        filt_df = generate_diff_cols(filt_df, req.metric_column, req.compare_column)

    unique_cols = set(not_cont_search_cols + [metric_col])
    updated_df = pd.concat([filt_df[list(unique_cols)], cont_df], axis=1).dropna()

    normalized_metric_col = np.array(updated_df[metric_col], dtype=float)
    # Invert metric column if ascending.
    metric_max = np.max(normalized_metric_col)
    if req.order_by == "ascending":
        normalized_metric_col = metric_max - normalized_metric_col

    cont_search_cols = [col + "_encode" for col in cont_search_cols]
    search_cols_str = not_cont_search_cols + cont_search_cols
    slice_finder = Slicefinder(alpha=req.alpha, k=20, max_l=req.max_lattice)
    slice_finder.fit(updated_df[search_cols_str].to_numpy(), normalized_metric_col)

    if slice_finder.top_slices_ is None or slice_finder.top_slices_statistics_ is None:
        return SliceFinderReturn(slices=[], metrics=[], sizes=[], overall_metric=0)

    discovered_slices: List[Slice] = []
    slice_metrics: List[float] = []
    slice_sizes: List[int] = []
    not_cont_search_num = len(not_cont_search_cols)

    for sli_i, sli in enumerate(slice_finder.top_slices_):
        # Rescale back to original metric.
        if req.order_by == "ascending":
            slice_metrics.append(
                metric_max
                - slice_finder.top_slices_statistics_[sli_i]["slice_average_error"]
            )
        else:
            slice_metrics.append(
                slice_finder.top_slices_statistics_[sli_i]["slice_average_error"]
            )

        slice_sizes.append(slice_finder.top_slices_statistics_[sli_i]["slice_size"])

        predicate_list = []
        for pred_i, sli_predicate in enumerate(sli):
            if sli_predicate is not None:
                join_val = "" if len(predicate_list) == 0 else "&"
                col = search_cols[pred_i]

                # not continuous columns
                if pred_i < not_cont_search_num:
                    if str(sli_predicate) in ["True", "False"]:
                        sli_predicate = "true" if sli_predicate else "false"
                    predicate_list.append(
                        FilterPredicate(
                            column=col,
                            operation="==",
                            value=sli_predicate,
                            join=join_val,
                        )
                    )
                # continuous columns
                else:
                    left_pred = FilterPredicate(
                        column=col,
                        operation=">=",
                        value=sli_predicate.left,
                        join="",
                    )
                    right_pred = FilterPredicate(
                        column=col,
                        operation="<",
                        value=sli_predicate.right,
                        join="&",
                    )
                    predicate_list.append(
                        FilterPredicateGroup(
                            predicates=[left_pred, right_pred], join=join_val
                        ),
                    )

        discovered_slices.append(
            Slice(
                slice_name="Generated Slice " + secrets.token_hex(nbytes=4),
                folder="",
                filter_predicates=FilterPredicateGroup(
                    predicates=predicate_list, join=""
                ),
            )
        )
    return SliceFinderReturn(
        slices=discovered_slices,
        metrics=slice_metrics,
        sizes=slice_sizes,
        overall_metric=slice_finder.average_error_
        if slice_finder.average_error_
        else 0,
    )



from typing import List

import numpy as np
import pandas as pd
from sklearn import preprocessing

from zeno.classes.base import MetadataType, ZenoColumn, ZenoColumnType
from zeno.classes.projection import Points2D, PointsColors


def check_embed_exists(df: pd.DataFrame, model: str):
    """Checks for the existence of an embedding column.
    Returns True if the column exists, False otherwise
    """
    embed_column = ZenoColumn(
        column_type=ZenoColumnType.EMBEDDING, name="embedding", model=model
    )
    exists = str(embed_column) in df.columns
    return exists and not df[str(embed_column)].isna().any()


cached_projections = {}


def run_tsne(df: pd.DataFrame, model: str) -> np.ndarray:
    """Project embedding into 2D space using t-SNE.

    Args:
        df (pd.DataFrame): Zeno DataFrame.
        model (str): Name of which model to use for the embedding.

    Returns:
        np.ndarray: 2D projection of the embedding.
    """

    if model in cached_projections:
        return cached_projections[model]

    from openTSNE import TSNE

    embed_col = ZenoColumn(
        column_type=ZenoColumnType.EMBEDDING, name="embedding", model=model
    )

    embed = df[str(embed_col)].to_numpy()
    embed = np.stack(embed, axis=0)  # type: ignore

    all_available_processors = -1
    default_iterations = 400
    tsne = TSNE(n_jobs=all_available_processors, n_iter=default_iterations)
    projection = tsne.fit(embed)

    cached_projections[model] = projection

    return projection


def projection_colors(df: pd.DataFrame, column: ZenoColumn) -> PointsColors:
    """Get colors for a projection based on a column.

    Args:
            df (pd.DataFrame): Dataframe with all columns from Zeno.
            column (ZenoColumn): Column to use for coloring the projection.

    Returns:
            PointsColors: The color range, the unique values, and the metadata type.
    """
    series = df[str(column)]
    unique = series.unique()
    metadata_type = "nominal"
    color_range: List[int] = []
    if len(unique) == 2:
        metadata_type = "boolean"
    if len(unique) > 10:
        if column.metadata_type == MetadataType.CONTINUOUS:
            metadata_type = "continuous"
            color_range = (
                np.interp(series, (series.min(), series.max()), (0, 20))
                .astype(int)
                .tolist()
            )
            unique = np.array([series.min(), series.max()])
        else:
            metadata_type = "other"
            color_range = [0] * len(series)
    else:
        labels = preprocessing.LabelEncoder().fit_transform(series)
        if isinstance(labels, np.ndarray):
            color_range = labels.astype(int).tolist()
        else:
            color_range = [0] * len(series)

    return PointsColors(
        color=color_range, domain=unique.tolist(), data_type=metadata_type
    )


def project_into_2d(
    df: pd.DataFrame, id_column: ZenoColumn, model: str, column: ZenoColumn
) -> Points2D:
    """If the embedding exists, will use t-SNE to project into 2D."""

    points = Points2D(x=[], y=[], color=[], domain=[], opacity=[], data_type="", ids=[])

    # Can't project without an embedding
    if not check_embed_exists(df, model):
        return points

    projection = run_tsne(df, model)

    # extract points and ids from computed projection
    points.x = projection[:, 0].tolist()
    points.y = projection[:, 1].tolist()
    color_results = projection_colors(df, column)
    points.color = color_results.color
    points.domain = color_results.domain
    points.data_type = color_results.data_type
    points.ids = df[str(id_column)].to_list()

    return points



"""Entry to Zeno. Parses TOML file, starts server, and runs the pipeline."""

import sys
from typing import Dict, Union

import pkg_resources
import uvicorn
from multiprocess import Process  # type: ignore

from zeno.api import ZenoParameters
from zeno.backend import ZenoBackend
from zeno.server import get_server
from zeno.setup import setup_zeno
from zeno.util import is_notebook, read_config

# Global variable to hold the Zeno server process.
# This is used to kill the server when re-running in a notebook.
ZENO_SERVER_PROCESS = None


def command_line():
    if len(sys.argv) == 1 or sys.argv[1] == "-h" or sys.argv[1] == "--help":
        print(
            "\n \033[1mZeno\033[0m",
            pkg_resources.get_distribution("zenoml-next").version,
            " - Machine learning evaluation framework.",
            "\n\n",
            "\033[1mUSAGE \033[0m \n\t",
            "zeno [-h] [-v] <config.toml>",
            "\n\n",
            "\033[1mARGUMENTS \033[0m \n\t",
            "<config.toml>\t\tZeno configuration file.\n\n"
            "\033[1m GLOBAL OPTIONS \033[0m \n\t",
            "-h (--help)\t\tDisplay this help message.\n"
            "\t -v (--version)\t\tDisplay this application version.\n",
        )

        sys.exit(0)

    if len(sys.argv) != 2:
        print(
            "ERROR: Zeno take one argument, either a configuration TOML file"
            + " or the keyword 'init'. "
            + "{0} arguments were passed.",
            len(sys.argv),
        )
        sys.exit(1)

    if sys.argv[1] == "-v" or sys.argv[1] == "--version":
        print(pkg_resources.get_distribution("zenoml").version)
        sys.exit(0)

    if sys.argv[1] == "init" or sys.argv[1] == "i":
        setup_zeno()
    else:
        zeno(sys.argv[1])


def run_zeno(params: ZenoParameters):
    zeno = ZenoBackend(params)
    app = get_server(zeno)

    zeno.start_processing()

    print(
        "\n\033[1mZeno\033[0m running on http://{}:{}\n".format(
            params.host, params.port
        )
    )
    uvicorn.run(app, host=params.host, port=params.port, log_level="error")


def zeno(args: Union[str, ZenoParameters, Dict]):
    """Main entrypoint for Zeno. This is called directly by the user in a notebook or
    script, or called by the command_line function when run by CLI.

    Args:
        args (Union[str, ZenoParameters, Dict]): The configuration for Zeno.
        ZenoParameters or dict when called from Python, str if called from commandline.
    """

    params = read_config(args)

    if params.serve:
        global ZENO_SERVER_PROCESS
        if ZENO_SERVER_PROCESS is not None:
            ZENO_SERVER_PROCESS.terminate()

        ZENO_SERVER_PROCESS = Process(
            target=run_zeno,
            args=(params,),
        )
        ZENO_SERVER_PROCESS.start()

        if not is_notebook():
            ZENO_SERVER_PROCESS.join()
    else:
        zeno = ZenoBackend(params)
        return zeno



from enum import Enum
from typing import Optional

from pandas import Series
from pydantic import BaseModel, ConfigDict


def to_camel(string: str) -> str:
    components = string.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


class CamelModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True  # replaces allow_population_by_field_name=True
    )


class ZenoColumnType(str, Enum):
    METADATA = "METADATA"
    PREDISTILL = "PREDISTILL"
    OUTPUT = "OUTPUT"
    EMBEDDING = "EMBEDDING"
    POSTDISTILL = "POSTDISTILL"


class MetadataType(str, Enum):
    NOMINAL = "NOMINAL"
    CONTINUOUS = "CONTINUOUS"
    BOOLEAN = "BOOLEAN"
    DATETIME = "DATETIME"
    OTHER = "OTHER"


class ZenoColumn(CamelModel):
    column_type: ZenoColumnType
    name: str
    metadata_type: Optional[MetadataType] = MetadataType.OTHER
    model: Optional[str] = ""

    # Append the column type to the name unless it is a vanilla metadata column.
    # We don't save the metadata_type since we dynamically change it
    # and it's not identifying.
    def __str__(self):
        m = ""
        if self.model is not None:
            m = self.model
        t = ""
        if self.column_type is not ZenoColumnType.METADATA:
            t = self.column_type
        return t + self.name + m

    def __hash__(self):
        return hash(str(self))


class DataProcessingReturn(BaseModel):
    """Return type for data processing functions."""

    column: ZenoColumn
    output: Series

    model_config = ConfigDict(arbitrary_types_allowed=True)



from typing import List, Optional

from zeno.classes.base import CamelModel, ZenoColumn
from zeno.classes.slice import FilterIds, FilterPredicateGroup, Slice


class SliceFinderRequest(CamelModel):
    metric_column: ZenoColumn
    search_columns: List[ZenoColumn]
    order_by: str
    alpha: float
    max_lattice: int
    compare_column: Optional[ZenoColumn] = None
    filter_predicates: Optional[FilterPredicateGroup] = None
    tag_ids: Optional[FilterIds] = None
    filter_ids: Optional[FilterIds] = None
    tag_list: Optional[List[str]] = None


class SliceFinderReturn(CamelModel):
    slices: List[Slice]
    metrics: List[float]
    sizes: List[int]
    overall_metric: float
